import requests
from hjc236_covid_dashboard.config_handler import get_config_data, ConfigError
import logging
from flask import Markup

log_file_location = get_config_data()["log_file_path"]
logging.basicConfig(filename=log_file_location, level=logging.DEBUG, format="%(asctime)s %(message)s")


def news_API_request(covid_terms: str = "covid COVID-19 coronavirus") -> list[dict]:
    """Returns relevant current news articles from the News API based on the covid_terms argument"""
    # TODO handle connection errors
    api_key = get_config_data()["news_api_key"]
    endpoint = "https://newsapi.org/v2/everything?"
    lang = get_config_data()["news_language"]
    formatted_url = endpoint + f"q={covid_terms}" + f"&apiKey={api_key}" + "&sortBy=publishedAt" + f"&language={lang}"
    response = requests.get(formatted_url, timeout=10)
    news_data = response.json()

    if news_data["status"] == "error":
        if news_data["code"] == "apiKeyInvalid":
            raise ConfigError("Invalid News API key in configuration file")
        else:
            logging.error("Failed to get articles from News API")

    news_data = news_data["articles"]
    return news_data


def update_news(update_name: str, deleted_articles: list[dict] = None) -> None:
    """Updates the global webpage_news_articles list with new content from the News API"""
    covid_terms = get_config_data()["news_covid_terms"]
    news_data = news_API_request(covid_terms)

    # If a list of deleted articles has been passed, check all new articles and delete them if their titles match
    # This is to ensure deleted articles do not return on page refresh
    if deleted_articles is not None:
        for article_index, article_dictionary in enumerate(news_data):
            if article_dictionary["title"] in deleted_articles:
                news_data.pop(article_index)

    # webpage_news_articles is the global list of articles in main.py passed to the webpage
    global webpage_news_articles
    webpage_news_articles = format_news_data(news_data)


def format_news_data(news_articles: list[dict]) -> list[dict]:
    """Changes the given list of news article dictionaries and alters them to look more user-friendly on the webpage.

    For each article: Replaces 'content' values with 'description' values if they exist for the article,
    which are better for displaying a concise article description. Formats titles to become HTML hyperlinked to the
    article's URL with flask.Markup()
    """

    formatted_articles = news_articles

    for article in formatted_articles:
        # 'description' is formatted in a more user-friendly way so it is better to display this to the user,
        # but not all articles have it - if they don't, better to just show the content than nothing
        if article["description"] is not None:
            article["content"] = article["description"]

        # Add hyperlink leading to the article's URL to all titles
        url = article["url"]
        title = article["title"]
        formatted_title = f"<a href={url}>{title}</a>"
        article["title"] = Markup(formatted_title)

    return formatted_articles



import logging
import sched
import time
import os

from uk_covid19 import Cov19API

from hjc236_covid_dashboard.config_handler import get_config_data

log_file_location = get_config_data()["log_file_path"]
current_location = os.path.abspath(os.path.dirname(__file__))
path = os.path.abspath(os.path.join(current_location, log_file_location))
logging.basicConfig(filename=path, level=logging.DEBUG, format="%(asctime)s %(message)s")


def parse_csv_data(csv_filename: str) -> list[str]:
    """Returns a list of strings for the rows in a given CSV file.

    NOTE: This is a purely demonstrative function made to fit the specification, actual data handling is done with JSON
    """
    current_location = os.path.abspath(os.path.dirname( __file__ ))
    file_path = os.path.abspath(os.path.join(current_location, csv_filename))

    with open(file_path, 'r') as csv_file:
        csv_rows = csv_file.read().splitlines()

    return csv_rows


def process_covid_csv_data(covid_csv_data: list[str]) -> tuple[int, int, int]:
    """Returns the number of cases over the last 7 days, current number of hospital cases, and cumulative deaths from
    given COVID data

    NOTE: This is a purely demonstrative function made to fit the specification, actual data handling is done with
    covid_data_handler.process_covid_JSON_data()
    """

    # Split CSV rows to create a 2D array.
    for i in range(0, len(covid_csv_data)):
        covid_csv_data[i] = covid_csv_data[i].split(",")

    last7days_cases = 0
    for i in range(3, 10):
        last7days_cases += int(covid_csv_data[i][6])

    current_hospital_cases = int(covid_csv_data[1][5])
    total_deaths = int(covid_csv_data[14][4])

    return last7days_cases, current_hospital_cases, total_deaths


def make_API_call(filters: list[str], structure: dict) -> dict:
    """Returns a JSON object containing COVID data based on the given filters and structure"""
    data = Cov19API(filters=filters, structure=structure)
    result = data.get_json()

    return result


def convert_JSON_data(list_of_dictionaries: list[dict]) -> dict[dict]:
    """Converts the list of dictionaries returned from the API to a single dictionary, where an item's key is the
    date of the COVID data, and the value is a dictionary of the remaining metrics"""

    dictionary_of_dictionaries = {}
    for i in range(0, len(list_of_dictionaries)):
        old_dictionary = list_of_dictionaries[i]
        date = old_dictionary.pop("date")
        dictionary_of_dictionaries[date] = old_dictionary

    return dictionary_of_dictionaries


def process_covid_json_data(local_json: dict, national_json: dict) -> dict:
    """Returns a dictionary of specified metrics based on the JSON files of local and national COVID data"""

    deaths_total = None
    hospitalCases = None
    national_7day_infections = 0
    local_7day_infections = 0

    counter = 0
    skipped_first_day = False
    for date in national_json.keys():
        current_data = national_json[date]

        # For cumDeaths and hospitalCases, find the first non-empty cells and use these values
        if current_data["cumDeaths"] is not None and deaths_total is None:
            deaths_total = current_data["cumDeaths"]

        if current_data["hospitalCases"] is not None and hospitalCases is None:
            hospitalCases = current_data["hospitalCases"]

        # Add up all the non-empty rows of 'newCases' until we have 7 (ie a week's worth of data)
        if current_data["newCases"] is not None and counter < 7:
            # Skip first day of COVID data as it is incomplete
            if skipped_first_day:
                national_7day_infections += current_data["newCases"]
                counter += 1
            else:
                skipped_first_day = True

    counter = 0
    skipped_first_day = False
    for date in local_json.keys():
        current_data = local_json[date]

        if current_data["newCases"] is not None and counter < 7:
            if skipped_first_day:
                local_7day_infections += current_data["newCases"]
                counter += 1
            else:
                skipped_first_day = True

    covid_data = {
        "local_7day_infections": local_7day_infections,  # local case total in the last 7 days

        "national_7day_infections": national_7day_infections,  # national case total in the last 7 days
        "hospitalCases": hospitalCases,  # current amount of hospitalised cases
        "deaths_total": deaths_total,  # current amount of cumulative deaths
    }

    return covid_data


def covid_API_request(location: str = "exeter", location_type: str = "ltla") -> dict:
    """Returns a dictionary of up-to-date COVID data via Public Health England's API"""

    nation_location = get_config_data()["nation_location"]

    national_filters = ["areaType=nation", f"areaName={nation_location}"]
    local_filters = [f"areaType={location_type}", f"areaName={location}"]

    local_structure = {"date": "date", "areaType": "areaType", "areaName": "areaName",
                       "newCases": "newCasesBySpecimenDate"}

    national_structure = {"date": "date", "areaType": "areaType", "areaName": "areaName",
                          "newCases": "newCasesByPublishDate", "hospitalCases": "hospitalCases",
                          "cumDeaths": "cumDailyNsoDeathsByDeathDate"}

    local_data = make_API_call(local_filters, local_structure)
    national_data = make_API_call(national_filters, national_structure)

    # strip headers from JSON data and convert from lists of dictionaries to one dictionary where key = date and
    # value = dictionary of remaining data
    local_data = convert_JSON_data(local_data["data"])
    national_data = convert_JSON_data(national_data["data"])

    processed_covid_data = process_covid_json_data(local_data, national_data)

    # Append location details to the dictionary here so the website can use them
    processed_covid_data["location"] = location
    processed_covid_data["nation_location"] = nation_location

    return processed_covid_data


def schedule_covid_updates(update_interval: int, update_name: str) -> None:
    """Updates COVID data every update_interval seconds in an update called update_name

    NOTE: This is a purely demonstrative function made to fit the specification, all updates are scheduled by the
    function web_interface.schedule_update()"""

    scheduler = sched.scheduler(time.time, time.sleep)

    # Wrapper function that can re-add itself to the scheduler each time it's run to ensure it repeats
    def scheduled_data_update() -> None:
        # Get local location and type from config file
        location = get_config_data()["local_location"]
        location_type = get_config_data()["local_location_type"]

        # Writes to a global variable for COVID data so it's updated whenever the scheduled API request happens
        global covid_data
        covid_data = covid_API_request(location, location_type)
        # Adds itself to the scheduler again when called so that it loops every update_interval seconds
        scheduler.enter(update_interval, 1, scheduled_data_update)

    # Add the event to the scheduler for the first time; after this it will keep adding itself again when called
    scheduler.enter(update_interval, 1, scheduled_data_update)
    scheduler.run(blocking=False)


def update_covid(update_name: str) -> None:
    """Updates the global webpage_covid_data list, this is in main.py and is what gets passed to the
    web page"""

    global webpage_covid_data
    location = get_config_data()["local_location"]
    location_type = get_config_data()["local_location_type"]

    webpage_covid_data = covid_API_request(location, location_type)





import time


def minutes_to_seconds(minutes: str) -> int:
    """Converts minutes to seconds"""
    return int(minutes) * 60


def hours_to_minutes(hours: str) -> int:
    """Converts hours to minutes"""
    return int(hours) * 60


def hhmm_to_seconds(hhmm: str) -> int:
    """Converts a string in the format hh:mm to a time in seconds from 00:00"""
    if len(hhmm.split(':')) != 2:
        print('Incorrect format. Argument must be formatted as HH:MM')
        return None
    return minutes_to_seconds(hours_to_minutes(hhmm.split(':')[0])) + \
           minutes_to_seconds(hhmm.split(':')[1])


def current_time_seconds() -> int:
    """Returns the current time from 00:00 in seconds"""
    hhmm = str(time.gmtime().tm_hour) + ":" + str(time.gmtime().tm_min)
    seconds = hhmm_to_seconds(hhmm)
    return seconds



import json
import os
import sys


def get_config_data() -> dict:
    """Returns the values from the config file as a dictionary"""
    current_location = os.path.abspath(os.path.dirname(__file__))
    path = os.path.abspath(os.path.join(current_location, 'config.json'))

    with open(path, 'r') as config_file:
        data = json.load(config_file)

    return data


class ConfigError(Exception):
    """Exception type for incorrect config file values."""
    pass


def validate_config_data(config_data: dict) -> None:
    """Checks that values in the config file are valid - if not, raise an error. This prevents the program running
    when a user has entered something invalid in the config file."""

    if config_data["news_api_key"] == "":
        raise ConfigError("news_api_key has not been set in config file")

    if config_data["news_language"] not in ['ar', 'de', 'en', 'es', 'fr', 'he', 'it', 'nl', 'no', 'pt', 'ru',
                                            'se', 'ud', 'zh']:
        raise ConfigError(f"{config_data['news_language']} in config file is not a valid option for news_language")

    if int(config_data["number_of_articles_to_display"]) < 0:
        raise ConfigError("number_of_articles_to_display in config file must be at least 0")

    if config_data["local_location_type"] not in ["ltla", "utla", "overview", "nation", "region", "nhsRegion"]:
        raise ConfigError("local_location_type in config file is invalid")

    if config_data["nation_location"] not in ["england", "northern ireland", "scotland", "wales"]:
        raise ConfigError("nation_location in config file is invalid")

    if int(config_data["repeat_interval_seconds"]) <= 0:
        raise ConfigError("repeat_interval_seconds in config file must be greater than 0")
